/**
 *	@file	devHwErr5565.c
 *	@brief	Device support for handling of hardware error of RM5565
 *	$Date:: 2022-11-10 15:05:32 +0900#$
 *	$Revision: 12 $
 *	$Author: iitsuka $
 *
 *	@page devsup_hwerr Device Support for Handling of error that was detected by hardware
 *	エラー監視用デバイスサポート
 *
 *	このデバイスサポートは、longin レコードタイプと組み合わせて使用する事で、
 *	リフレクティブメモリのエラー監視を行うためのものである。@n
 *	Device Type 名は @c "H/W error"。
 *
 *	@par Synopsis
 *	@code{.unparsed}
 *	record(longin, "pv-name") {
 *		field(DTYP, "H/W error")
 *		field(SCAN, "I/O Intr")
 *		field(INP,  "#C<card_number> S @<error_type>[,<severity>]")
 *	}
 *	@endcode
 *	@b card_number @n
 *	@li The card number designated in argument @e card in @ref RM5565Configure
 *
 *	@b error_type
 *	@li @c BadData ... Bad Data
 *	@li @c FIFOFull ... RX FIFO Full
 *	@li @c RogueDet ... Rogue Packet Detected and Removed
 *	@li @c FIFOAFull ... RX FIFO Almost Full
 *	@li @c SyncLoss ... Sync Loss
 *
 *	@b severity
 *	@li @c NO_ALARM ... NO_ALARM
 *	@li @c MINOR ... MINOR_ALARM
 *	@li @c MAJOR ... MAJOR_ALARM
 *	@li @c INVALID ... INVALID_ALARM
 *
 *	@par Usage
 *	@code{.unparsed}
 *	record(longin, "RM5565:DEV0:BadData") {
 *		field(DTYP, "H/W error")
 *		field(SCAN, "I/O Intr")
 *		field(INP,  "#C0 S @BadData,MINOR")
 *			・
 *			・
 *			・
 *	}
 *	@endcode
 *	@c INP フィールドでカード番号/エラータイプを指定する。@n
 *	@c "C" に続く数字(例では @c '0')でカード番号を、
 *	@c "@" に続く文字列(例では @c 'BadData')でエラータイプを指定しておく。@n
 *	また、',' に続けて Severity (例では @c 'MINOR') を指定する事も可能である。
 *	Severity 指定は省略も可能で、省略されると @c 'NO_ALARM' が指定されたのと同様の動作となる。@n
 *	指定されたエラーが発生すると、@c VAL が +1 されてレコードがプロセスされる。@n
 *	この時、 @c INP での Severity 指定が NO_ALARM 以外だった場合、
 *	レコードは指定された Severity で COMM_ALARM 状態となる。@n
 *	発生したエラーが解除されると、自動でレコードの Alarm 状態が解除されるが、
 *	@c VAL の値は保持される。@n
 *	つまり、IOC 起動後からの累積発生回数を保持するという事である。@n
 *	@code
 *	$ camonitor RM5565:DEV0:BadData
 *	RM5565:DEV0:BadData   2016-08-30 13:15:52.958786 0
 *	RM5565:DEV0:BadData   2016-08-30 13:17:32.982779 1 COMM MINOR   ← Bad Data エラー発生
 *	RM5565:DEV0:BadData   2016-08-30 13:17:37.674690 1              ← Bad Data エラー解除
 *	@endcode
 */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <alarm.h>
#include <dbFldTypes.h>
#include <dbAccess.h>
#include <dbScan.h>
#include <recGbl.h>
#include <devSup.h>
#include <devLib.h>
#include <drvSup.h>
#include <menuAlarmSevr.h>
#include <epicsExport.h>
#include <epicsThread.h>
#include <epicsMutex.h>
#include <iocsh.h>
#include <epicsString.h>
#include <epicsInterrupt.h>
#include <epicsAssert.h>
#include <cantProceed.h>
#include <errlog.h>

#include <longinRecord.h>

#include "drv5565.h"

/**	RM5565 H/W error type
 */
typedef	enum {
	HWERR_BADDATA = 0,	/**< Bad Data				*/
	HWERR_FIFOFULL,		/**< RX FIFO Full			*/
	HWERR_ROGUEDET,		/**< Rogue Packet Detect	*/
	HWERR_FIFOAFULL,	/**< RX FIFO Almost Full	*/
	HWERR_SYNCLOSS,		/**< Sync Loss				*/
	HWERR_UNKNOWN		/**< Unknown				*/
} hwerr_type_t;

/**	Device Support Private Information
 */
typedef	struct {
	RM5565_t handle;		/**< handle of specific RM5565	*/
	hwerr_type_t errType;	/**< H/W error type				*/
	menuAlarmSevr sevr;		/**< Alarm Severity				*/
} hwerr_dpvt_t;

/**	convert string to Hardware Error Type enumerator
 *	@param[in]	param	string
 *	@return	H/W error type enumerator
 */
static hwerr_type_t hwerr_type(const char *param)
{
	if ((epicsStrCaseCmp(param, "BadData") == 0)
	  || (epicsStrCaseCmp(param, "Bad Data") == 0))
	{
		return HWERR_BADDATA;
	}
	else if ((epicsStrCaseCmp(param, "FIFOFull") == 0)
	  || (epicsStrCaseCmp(param, "FIFO Full") == 0))
	{
		return HWERR_FIFOFULL;
	}
	else if ((epicsStrCaseCmp(param, "RogueDet") == 0)
	  || (epicsStrCaseCmp(param, "Rogue Det") == 0))
	{
		return HWERR_ROGUEDET;
	}
	else if ((epicsStrCaseCmp(param, "FIFOAFull") == 0)
	  || (epicsStrCaseCmp(param, "FIFO AFull") == 0))
	{
		return HWERR_FIFOAFULL;
	}
	else if ((epicsStrCaseCmp(param, "SyncLoss") == 0)
	  || (epicsStrCaseCmp(param, "Sync Loss") == 0))
	{
		return HWERR_SYNCLOSS;
	}
	return HWERR_UNKNOWN;
}

/**	convert string to Alarm Severity
 *	@param[in]	param	string
 *	@return	Alarm Severity
 */
static menuAlarmSevr sevr_type(const char *param)
{
	if (param == NULL)
	{
		return NO_ALARM;
	}
	else if (epicsStrCaseCmp(param, "MINOR") == 0)
	{
		return MINOR_ALARM;
	}
	else if (epicsStrCaseCmp(param, "MAJOR") == 0)
	{
		return MAJOR_ALARM;
	}
	else if (epicsStrCaseCmp(param, "INVALID") == 0)
	{
		return INVALID_ALARM;
	}
	else
	{
		return NO_ALARM;
	}
}

/**	initialize specific record
 *	@param[in,out]	prec	pointer to record information
 *	@retval	0					OK
 *	@retval	S_dev_badInpType	bad INP field type
 *	@retval	S_dev_badCard		bad card number
 *	@retval	S_dev_noMemory		malloc() failed
 *	@retval	S_db_badField		the parameter that was described in INP field is invalid
 */
static long li_hwerr_init_record(struct longinRecord *prec)
{
	long status;

	if (prec->inp.type != VME_IO)
	{
		recGblRecordError(status = S_dev_badInpType, prec, __func__);
	}
	else
	{
		hwerr_dpvt_t *dpvt;
		char *pParmDup;

		dpvt = (hwerr_dpvt_t *)mallocMustSucceed(sizeof(hwerr_dpvt_t), __func__);
		dpvt->handle = drvRM5565_getHandle(prec->inp.value.vmeio.card);
		if (dpvt->handle == NULL)
		{
			recGblRecordError(status = S_dev_badCard, prec, __func__);
		}
		else if ((pParmDup = strdup(prec->inp.value.vmeio.parm)) == NULL)
		{
			recGblRecordError(status = S_dev_noMemory, prec, __func__);
		}
		else
		{
			char *pTmp, *savep;

			pTmp = strtok_r(pParmDup, ",", &savep);
			if (pTmp == NULL)
			{
				recGblRecordError(status = S_db_badField, prec, __func__);
			}
			else if ((dpvt->errType = hwerr_type(pTmp)) == HWERR_UNKNOWN)
			{
				recGblRecordError(status = S_db_badField, prec, __func__);
			}
			else
			{
				dpvt->sevr = sevr_type(strtok_r(NULL, "", &savep));
				prec->dpvt = dpvt;
				status = 0L;
			}
			free(pParmDup);
		}
		if (status != 0L)
		{
			free(dpvt);
		}
	}
	return status;
}

/**	get I/O interrupt information
 *	@param[in]		cmd		command
 *							- @c 0 : add to
 *							- @c 1 : delete from
 *	@param[in,out]	prec	pointer to record information
 *	@param[out]		ppvt	pointer to a variable where IOSCANPVT will be stored
 *	@retval	0				OK
 *	@retval	S_dev_NoInit	NG
 */
static long li_hwerr_get_ioint_info(int cmd, struct longinRecord *prec, IOSCANPVT *ppvt)
{
	long status;

	if (prec->dpvt == NULL)
	{
		recGblRecordError(status = S_dev_NoInit, prec, __func__);
	}
	else
	{
		hwerr_dpvt_t *dpvt = (hwerr_dpvt_t *)prec->dpvt;

		switch (dpvt->errType)
		{
		case HWERR_BADDATA:
			*ppvt = drvRM5565_info(dpvt->handle)->scan_baddata;
			break;

		case HWERR_FIFOFULL:
			*ppvt = drvRM5565_info(dpvt->handle)->scan_fifofull;
			break;

		case HWERR_ROGUEDET:
			*ppvt = drvRM5565_info(dpvt->handle)->scan_roguedet;
			break;

		case HWERR_FIFOAFULL:
			*ppvt = drvRM5565_info(dpvt->handle)->scan_fifoafull;
			break;

		case HWERR_SYNCLOSS:
			*ppvt = drvRM5565_info(dpvt->handle)->scan_syncloss;
			break;

		default:
			cantProceed("devHwErr5565: unknown error type(=%d)\n", dpvt->errType);
			break;
		}
		status = 0L;
	}
	return status;
}

/**	read record
 *	@param[in,out]	prec	pointer to record information
 *	@retval	0				OK
 *	@retval	S_dev_NoInit	NG
 */
static long li_hwerr_read_record(struct longinRecord *prec)
{
	long status;

	if (prec->dpvt == NULL)
	{
		recGblRecordError(status = S_dev_NoInit, prec, __func__);
	}
	else
	{
		hwerr_dpvt_t *dpvt = (hwerr_dpvt_t *)prec->dpvt;

		switch (dpvt->errType)
		{
		case HWERR_BADDATA:
			prec->val = drvRM5565_info(dpvt->handle)->count_baddata;
			break;

		case HWERR_FIFOFULL:
			prec->val = drvRM5565_info(dpvt->handle)->count_fifofull;
			break;

		case HWERR_ROGUEDET:
			prec->val = drvRM5565_info(dpvt->handle)->count_roguedet;
			break;

		case HWERR_FIFOAFULL:
			prec->val = drvRM5565_info(dpvt->handle)->count_fifoafull;
			break;

		case HWERR_SYNCLOSS:
			prec->val = drvRM5565_info(dpvt->handle)->count_syncloss;
			break;

		default:
			cantProceed("devHwErr5565: unknown error type(=%d)\n", dpvt->errType);
			break;
		}
		if (prec->val != prec->lalm)
		{
			recGblSetSevr(prec, COMM_ALARM, dpvt->sevr);
		}
		status = 0L;
	}
	return status;
}

/**	@cond		*/
static struct {
	long number;
	DEVSUPFUN report;
	DEVSUPFUN init;
	DEVSUPFUN init_record;
	DEVSUPFUN get_ioint_info;
	DEVSUPFUN read_record;
} devLiHwerr = {
	5L,
	NULL,
	NULL,
	li_hwerr_init_record,
	li_hwerr_get_ioint_info,
	li_hwerr_read_record
};
epicsExportAddress(dset, devLiHwerr);
/**	@endcond	*/

