/**
 *	@file	drv5565pci.c
 *	@brief	EPICS driver support for PCIE-5565
 *	$Date:: 2022-11-10 13:40:09 +0900#$
 *	$Revision: 9 $
 *	$Author: iitsuka $
 */
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>

#include <alarm.h>
#include <dbFldTypes.h>
#include <dbScan.h>
#include <recGbl.h>
#include <devSup.h>
#include <devLib.h>
#include <drvSup.h>
#include <epicsTypes.h>
#include <epicsExport.h>
#include <epicsThread.h>
#include <epicsMutex.h>
#include <epicsExit.h>
#include <ellLib.h>
#include <iocsh.h>
#include <initHooks.h>
#include <epicsString.h>
#include <epicsRingBytes.h>
#include <epicsAssert.h>
#include <cantProceed.h>
#include <errlog.h>
#include <callback.h>
#include <rfm2g_api.h>

#include "drv5565.h"

#define	countof(a)	(sizeof(a)/sizeof(a[0u]))

#define	NUM_NINT	(4)		/**< number of network interrupt	*/

#define	ERR(api, result) \
	(void)errlogPrintf(__FILE__":%d " #api "() failed. %s\n", __LINE__, RFM2gErrorMsg(result));

#define	API(api, ...) \
	do { \
		RFM2G_STATUS result; \
		result = api(__VA_ARGS__); \
		if (result != RFM2G_SUCCESS) \
		{ \
			ERR(api, result); \
		} \
	} while (0)

/**
 *	structure of mapped memory
 */
typedef struct {
	ELLNODE link;			/**< information of node link			*/
	volatile void *memptr;	/**< memory pointer that was assigned	*/
	epicsUInt32 offset;		/**< designated offset					*/
	epicsUInt32 size;		/**< designated size(in bytes)			*/
} mapinfo_t;

/**
 *	structure of device information
 */
typedef struct {
	ELLNODE link;					/**< information of node link			*/
	IOSCANPVT scan_nint[NUM_NINT];	/**< IOSCANPVT for network interrupt	*/
	IOSCANPVT scan_baddata;			/**< IOSCANPVT for Bad Data				*/
	IOSCANPVT scan_fifofull;		/**< IOSCANPVT for RX FIFO Full			*/
	IOSCANPVT scan_roguedet;		/**< IOSCANPVT for Rogue pkt Detected	*/
	IOSCANPVT scan_fifoafull;		/**< IOSCANPVT for RX FIFO Almost Full	*/
	IOSCANPVT scan_syncloss;		/**< IOSCANPVT for Sync Loss			*/
	epicsUInt32 count_baddata;		/**< counter of Bad Data				*/
	epicsUInt32 count_fifofull;		/**< counter of RX FIFO Full			*/
	epicsUInt32 count_roguedet;		/**< counter of Rogue Pkt Detected		*/
	epicsUInt32 count_fifoafull;	/**< counter of RX FIFO Almost Full		*/
	epicsUInt32 count_syncloss;		/**< counter of Sync Loss				*/
	int card;						/**< card number						*/
	RFM2GHANDLE handle;				/**< rfm2g handle						*/
	RFM2G_UINT32 offset_sta;		/**< start of offset					*/
	RFM2G_UINT32 offset_end;		/**< end of offset						*/
	ELLLIST maplist;				/**< list of mapped memory				*/
	epicsRingBytesId sid[NUM_NINT];	/**< FIFO of node number				*/
	epicsRingBytesId isd[NUM_NINT];	/**< FIFO of extended data				*/
	CALLBACK poll_baddata;			/**< callback of BAD DATA				*/
	CALLBACK poll_syncloss;			/**< callback of SYNC LOSS				*/
	char devpath[64u];				/**< device path						*/
} devinfo_t;

static ELLLIST l_list = ELLLIST_INIT;

/**	find device information
 *	@param[in]	card	card number
 *	@return	pointer to device information structure
 */
static devinfo_t *find_dev(int card)
{
	devinfo_t *pdev;

	pdev = (devinfo_t *)ellFirst(&l_list);
	while (pdev != NULL)
	{
		if (pdev->card == card)
		{
			break;
		}
		pdev = (devinfo_t *)ellNext(&pdev->link);
	}
	return pdev;
}

/**	create FIFOs of network-interrupt
 *	@param[in,out]	pdev	pointer to device information
 */
static int create_fifos(devinfo_t *pdev)
{
	size_t i;

	for (i = 0u; i < countof(pdev->sid); i++)
	{
		pdev->sid[i] = epicsRingBytesCreate(sizeof(epicsUInt8) * 256u);
		if (pdev->sid[i] == NULL)
		{
			return -1;
		}
	}
	for (i = 0u; i < countof(pdev->isd); i++)
	{
		pdev->isd[i] = epicsRingBytesCreate(sizeof(epicsUInt32) * 256u);
		if (pdev->isd[i] == NULL)
		{
			return -1;
		}
	}
	return 0;
}

/**	destroy FIFOs of network-interrupt
 *	@param[in]	pdev	pointer to device information
 */
static void delete_fifos(devinfo_t *pdev)
{
	size_t i;

	for (i = 0u; i < countof(pdev->sid); i++)
	{
		if (pdev->sid[i] != NULL)
		{
			epicsRingBytesDelete(pdev->sid[i]);
			pdev->sid[i] = NULL;
		}
	}
	for (i = 0u; i < countof(pdev->isd); i++)
	{
		if (pdev->isd[i] != NULL)
		{
			epicsRingBytesDelete(pdev->isd[i]);
			pdev->isd[i] = NULL;
		}
	}
}

/**	enable all event callbacks
 *	@param[in]	pdev		pointer to device information
 *	@param[in]	callback	event callback function
 */
static void enable_callback(devinfo_t *pdev, RFM2G_EVENT_FUNCPTR callback)
{
	static const RFM2GEVENTTYPE s_type[] = {
		RFM2GEVENT_INTR1,
		RFM2GEVENT_INTR2,
		RFM2GEVENT_INTR3,
		RFM2GEVENT_INTR4,
		RFM2GEVENT_BAD_DATA,
		RFM2GEVENT_RXFIFO_FULL,
		RFM2GEVENT_ROGUE_PKT,
		RFM2GEVENT_RXFIFO_AFULL,
		RFM2GEVENT_SYNC_LOSS
	};
	size_t i;

	for (i = 0u; i < countof(s_type); i++)
	{
		API(RFM2gClearEvent, pdev->handle, s_type[i]);
		API(RFM2gEnableEvent, pdev->handle, s_type[i]);
	}
	for (i = 0u; i < countof(s_type); i++)
	{
		API(RFM2gEnableEventCallback, pdev->handle, s_type[i], callback);
	}
}

/**	disable all event callbacks
 *	@param[in]	pdev	pointer to device information
 */
static void disable_callback(devinfo_t *pdev)
{
	static const RFM2GEVENTTYPE s_type[] = {
		RFM2GEVENT_SYNC_LOSS,
		RFM2GEVENT_RXFIFO_AFULL,
		RFM2GEVENT_ROGUE_PKT,
		RFM2GEVENT_RXFIFO_FULL,
		RFM2GEVENT_BAD_DATA,
		RFM2GEVENT_INTR4,
		RFM2GEVENT_INTR3,
		RFM2GEVENT_INTR2,
		RFM2GEVENT_INTR1
	};
	size_t i;

	for (i = 0u; i < countof(s_type); i++)
	{
		API(RFM2gDisableEventCallback, pdev->handle, s_type[i]);
	}
}

/**	unmap all mapped memories
 *	@param[in,out]	pdev	pointer to device information
 */
static void unmap_all(devinfo_t *pdev)
{
	mapinfo_t *pmap;

	pmap = (mapinfo_t *)ellLast(&pdev->maplist);
	while (pmap != NULL)
	{
		mapinfo_t *pmapPrev = (mapinfo_t *)ellPrevious(&pmap->link);

		API(RFM2gUnMapUserMemoryBytes, pdev->handle, &pmap->memptr, pmap->size);
		ellDelete(&pdev->maplist, &pmap->link);
		free(pmap);
		pmap = pmapPrev;
	}
}

/**	get device handle
 *	@param[in]	card	card number
 *	@return	device handle
 */
RM5565_t drvRM5565_getHandle(int card)
{
	devinfo_t *pdev;

	pdev = find_dev(card);
	if (pdev == NULL)
	{
		return NULL;
	}
	return (RM5565_t)pdev;
}

/**	get pointer to RM5565 information
 *	@param[in]	handle	RM5565 handle
 *	@return	pointer to RM5565info_t structure
 */
inline RM5565info_t *drvRM5565_info(RM5565_t handle)
{
	return (RM5565info_t *)handle;
}

/**	validity check of offset
 *	@param[in]	handle	RM5565 handle
 *	@param[in]	offset	offset
 *	@retval	true	offset is valid
 *	@retval	false	offset is invalid
 */
inline bool drvRM5565_validOffset(RM5565_t handle, ptrdiff_t offset)
{
	bool status;

	if (handle == NULL)
	{
		status = false;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;

		if ((pdev->offset_sta <= offset) && (offset <= pdev->offset_end))
		{
			status = true;
		}
		else
		{
			status = false;
		}
	}
	return status;
}

#if		defined(vxWorks)
/**	address of reflective memory
 *	@param[in]	handle	RM5565 handle
 *	@param[in]	offset	offset
 *	@return	pointer to RM5565 memory
 */
inline volatile void *drvRM5565_ptr(RM5565_t handle, ptrdiff_t offset)
{
	return (volatile void *)(offset);
}
#endif	/* defined(vxWorks) */

/**	read SID data(source node ID)
 *	@param[in]	handle	RM5565 handle
 *	@param[in]	intId	Network-Interrupt ID
 *						- @c 0 : Type 1 Network-Interrupt
 *						- @c 1 : Type 2 Network-Interrupt
 *						- @c 2 : Type 3 Network-Interrupt
 *						- @c 3 : Type 4 Network-Interrupt
 *	@param[out]	pSID	pointer to a variable where SID will be stored
 *	@retval	0	OK
 *	@retval	-1	data is not ready
 *	@retval	-2	argument error
 */
int drvRM5565_readSID(RM5565_t handle, epicsUInt8 intId, epicsUInt8 *pSID)
{
	int status;

	if ((handle == NULL) || (pSID == NULL) || (intId > 3u))
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		epicsUInt8 sid;

		if (epicsRingBytesGet(pdev->sid[intId], (char *)&sid, sizeof(sid)) != sizeof(sid))
		{
			status = -1;
		}
		else
		{
			*pSID = sid;
			status = 0;
		}
	}
	return status;
}

/**	read ISD data(Network-Interrupt data)
 *	@param[in]	handle	RM5565 handle
 *	@param[in]	intId	Network-Interrupt ID
 *						- @c 0 : Type 1 Network-Interrupt
 *						- @c 1 : Type 2 Network-Interrupt
 *						- @c 2 : Type 3 Network-Interrupt
 *						- @c 3 : Type 4 Network-Interrupt
 *	@param[out]	pISD	pointer to a variable where ISD will be stored
 *	@retval	0	OK
 *	@retval	-1	data is not ready
 *	@retval	-2	argument error
 */
int drvRM5565_readISD(RM5565_t handle, epicsUInt8 intId, epicsUInt32 *pISD)
{
	int status;

	if ((handle == NULL) || (pISD == NULL) || (intId > 3u))
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		epicsUInt32 isd;

		if (epicsRingBytesGet(pdev->isd[intId], (char *)&isd, sizeof(isd)) != sizeof(isd))
		{
			status = -1;
		}
		else
		{
			*pISD = isd;
			status = 0;
		}
	}
	return status;
}

/**	issues Network-Interrupt command
 *	@param[in]	handle	RM5565 handle
 *	@param[in]	node	Target Node ID
 *	@param[in]	intId	Network-Interrupt ID
 *						- @c 0 : Type 1 Network-Interrupt
 *						- @c 1 : Type 2 Network-Interrupt
 *						- @c 2 : Type 3 Network-Interrupt
 *						- @c 3 : Type 4 Network-Interrupt
 *	@param[in]	data	Network-Interrupt data
 *	@retval	0	OK
 *	@retval	-2	argument error
 *
 *	- - -
 *	引数 @e node に対して、引数 @e intId で指定した Network-Interrupt を送信する。@n
 *	引数 @e handle には、@ref drvRM5565_getHandle で取得したハンドルを指定する。@n
 *	引数 @e node に0～255の値を指定すると、対応するノードのみに Network-Interrupt 
 *	を送信するが、256以上の値を指定すると、接続されている全ノードに対して、
 *	Network-Interruptを送信できる。@n
 *
 *	@see RM5565_t, drvRM5565_getHandle
 */
int drvRM5565_txInt(RM5565_t handle, epicsUInt16 node, epicsUInt8 intId, epicsUInt32 data)
{
	static const RFM2GEVENTTYPE s_type[NUM_NINT] = {
		RFM2GEVENT_INTR1,
		RFM2GEVENT_INTR2,
		RFM2GEVENT_INTR3,
		RFM2GEVENT_INTR4
	};
	int status;

	if ((handle == NULL) || (intId >= countof(s_type)))
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		RFM2G_NODE nid;
		RFM2G_STATUS result;

		nid = (node > 255u) ? RFM2G_NODE_ALL : (RFM2G_NODE)node;
		result = RFM2gSendEvent(pdev->handle, nid, s_type[intId], data);
		if (result != RFM2G_SUCCESS)
		{
			ERR(RFM2gSendEvent, result);
			status = -1;
		}
		else
		{
			status = 0;
		}
	}
	return status;
}

/**	clearing of OWN_DATA bit
 *	@param[in]	handle	RM5565 handle
 *	@retval	0	OK
 *	@retval	-2	argument error
 */
int drvRM5565_clearOwnData(RM5565_t handle)
{
	int status;

	if (handle == NULL)
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		RFM2G_BOOL state;
		RFM2G_STATUS result;

		result = RFM2gClearOwnData(pdev->handle, &state);
		if (result != RFM2G_SUCCESS)
		{
			ERR(RFM2gClearOwnData, result);
			status = -1;
		}
		else
		{
			status = 0;
		}
	}
	return status;
}

/**	waiting for set of OWN_DATA bit
 *	@param[in]	handle	RM5565 handle
 *	@param[in]	tmout	Timeout time(in seconds)
 *	@retval	0	OK
 *	@retval	-1	timeout
 *	@retval	-2	argument error
 */
int drvRM5565_waitOwnData(RM5565_t handle, epicsFloat64 tmout)
{
	int status;

	if (handle == NULL)
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else if (tmout <= 0.0)
	{
		status = 0;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		RFM2G_STATUS result;
		RFM2G_UINT32 lcsr;
		epicsFloat64 sleeptm = epicsThreadSleepQuantum();

		status = 0;

		result = RFM2gReadReg(pdev->handle, RFM2GCTRLREGMEM, rfm2g_lcsr, sizeof(lcsr), &lcsr);
		if (result != RFM2G_SUCCESS)
		{
			ERR(RFM2gReadReg, result);
			status = -1;
		}
		else
		{
			while ((lcsr & RFM2G_LCSR_OWNDATA) == 0u)
			{
				if ((tmout -= sleeptm) < 0.0)
				{
					status = -1;
					break;
				}
				epicsThreadSleep(sleeptm);
				result = RFM2gReadReg(pdev->handle, RFM2GCTRLREGMEM, rfm2g_lcsr, sizeof(lcsr), &lcsr);
				if (result != RFM2G_SUCCESS)
				{
					ERR(RFM2gReadReg, result);
					status = -1;
					break;
				}
			}
		}
	}
	return status;
}

/**	read LED status
 *	@param[in]	handle	handle of RM5565
 *	@retval	1	on
 *	@retval	0	off
 *	@retval	-1	error
 *	@retval	-2	argument error
 */
int drvRM5565_readLED(RM5565_t handle)
{
	int status;

	if (handle == NULL)
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		RFM2G_STATUS result;
		RFM2G_BOOL led;

		result = RFM2gGetLed(pdev->handle, &led);
		if (result != RFM2G_SUCCESS)
		{
			ERR(RFM2gGetLed, result);
			status = -1;
		}
		else
		{
			status = (led == RFM2G_ON) ? 1 : 0;
		}
	}
	return status;
}

/**	write LED condition
 *	@param[in]	handle	handle of RM5565
 *	@param[in]	cond	LED condition
 *						- @c 0 : off
 *						- @c positve : on
 *						- @c negative : complement
 *	@retval	0	last status is off
 *	@retval	1	last status is on
 *	@retval	-1	API error
 *	@retval	-2	argument error
 */
int drvRM5565_writeLED(RM5565_t handle, int cond)
{
	int status;

	if (handle == NULL)
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		status = drvRM5565_readLED(handle);
		if ((status == 0) || (status == 1))
		{
			if (cond < 0)
			{
				cond = (status == 0) ? 1 : 0;
			}
		}
		if (status >= 0)
		{
			devinfo_t *pdev = (devinfo_t *)handle;
			RFM2G_BOOL led = (cond == 0) ? RFM2G_FALSE : RFM2G_TRUE;
			RFM2G_STATUS result;

			result = RFM2gSetLed(pdev->handle, led);
			if (result != RFM2G_SUCCESS)
			{
				ERR(RFM2gSetLed, result);
				status = -1;
			}
		}
	}
	return status;
}

#if		defined(linux)
/**	
 *	@param[in]	handle	RM handle
 *	@param[in]	threshold	DMA threshold value
 *	@retval	0	OK
 *	@retval	-1	API error
 *	@retval	-2	invalid argument
 */
int drvRM5565_dmaThreshold(RM5565_t handle, epicsUInt32 threshold)
{
	int status;

	if (handle == NULL)
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		RFM2G_STATUS result;

		result = RFM2gSetDMAThreshold(pdev->handle, threshold);
		if (result != RFM2G_SUCCESS)
		{
			ERR(RFM2gSetDMAThreshold, result);
			status = -1;
		}
		else
		{
			status = 0;
		}
	}
	return status;
}

/**	
 *	@param[in]	handle	RM handle
 *	@param[in]	fswap	swap flag
 *						- @c true : swap on
 *						- @c false : swap off
 *	@retval	0	OK
 *	@retval	-1	API error
 *	@retval	-2	invalid argument
 */
int drvRM5565_dmaByteSwap(RM5565_t handle, bool fswap)
{
	int status;

	if (handle == NULL)
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		RFM2G_STATUS result;
		RFM2G_BOOL param;

		param = fswap ? RFM2G_TRUE : RFM2G_FALSE;
		result = RFM2gSetDMAByteSwap(pdev->handle, param);
		if (result != RFM2G_SUCCESS)
		{
			ERR(RFM2gSetDMAByteSwap, result);
			status = -1;
		}
		else
		{
			status = 0;
		}
	}
	return status;
}
#endif	/* defined(linux) */

/**	read form reflective memory
 *	@param[in]	handle	RM handle
 *	@param[in]	offset	offset
 *	@param[out]	pbuff	pointer to buffer area that stores result
 *	@param[in]	bytes	size of pbuff
 *	@retval	0	OK
 *	@retval	-1	API error
 *	@retval	-2	invalid argument
 */
int drvRM5565_read(RM5565_t handle, epicsUInt32 offset, void *pbuff, epicsUInt32 bytes)
{
	int status;

	if ((handle == NULL) || (pbuff == NULL))
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		RFM2G_STATUS result;

		result = RFM2gRead(pdev->handle, offset, pbuff, bytes);
		if (result != RFM2G_SUCCESS)
		{
			ERR(RFM2gRead, result);
			status = -1;
		}
		else
		{
			status = 0;
		}
	}
	return status;
}

/**	read 32-bit data from reflective memory
 *	@param[in]	handle	RM handle
 *	@param[in]	offset	offset
 *	@param[out]	pdata	pointer to 32bits variable that stores result
 *	@retval	0	OK
 *	@retval	-1	API error
 *	@retval	-2	invalid argument
 */
int drvRM5565_read32(RM5565_t handle, epicsUInt32 offset, void *pdata)
{
	int status;

	if ((handle == NULL) || (pdata == NULL))
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		RFM2G_UINT32 data;
		RFM2G_STATUS result;

		result = RFM2gPeek32(pdev->handle, offset, &data);
		if (result != RFM2G_SUCCESS)
		{
			ERR(RFM2gPeek32, result);
			status = -1;
		}
		else
		{
			*((epicsUInt32 *)pdata) = data;
			status = 0;
		}
	}
	return status;
}

/**	read 16-bit data from reflective memory
 *	@param[in]	handle	RM handle
 *	@param[in]	offset	offset
 *	@param[out]	pdata	pointer to 16bits variable that stores result
 *	@retval	0	OK
 *	@retval	-1	API error
 *	@retval	-2	invalid argument
 */
int drvRM5565_read16(RM5565_t handle, epicsUInt32 offset, void *pdata)
{
	int status;

	if ((handle == NULL) || (pdata == NULL))
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		RFM2G_UINT16 data;
		RFM2G_STATUS result;

		result = RFM2gPeek16(pdev->handle, offset, &data);
		if (result != RFM2G_SUCCESS)
		{
			ERR(RFM2gPeek16, result);
			status = -1;
		}
		else
		{
			*((epicsUInt16 *)pdata) = data;
			status = 0;
		}
	}
	return status;
}

/**	read 8-bit data from reflective memory
 *	@param[in]	handle	RM handle
 *	@param[in]	offset	offset
 *	@param[out]	pdata	pointer to 8bits variable that stores result
 *	@retval	0	OK
 *	@retval	-1	API error
 *	@retval	-2	invalid argument
 */
int drvRM5565_read8(RM5565_t handle, epicsUInt32 offset, void *pdata)
{
	int status;

	if ((handle == NULL) || (pdata == NULL))
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		RFM2G_UINT8 data;
		RFM2G_STATUS result;

		result = RFM2gPeek8(pdev->handle, offset, &data);
		if (result != RFM2G_SUCCESS)
		{
			ERR(RFM2gPeek8, result);
			status = -1;
		}
		else
		{
			*((epicsUInt8 *)pdata) = data;
			status = 0;
		}
	}
	return status;
}

/**	write to reflective memory
 *	@param[in]	handle	RM handle
 *	@param[in]	offset	offset
 *	@param[out]	pbuff	pointer to buffer area
 *	@param[in]	bytes	size of pbuff
 *	@retval	0	OK
 *	@retval	-1	API error
 *	@retval	-2	invalid argument
 */
int drvRM5565_write(RM5565_t handle, epicsUInt32 offset, void *pbuff, epicsUInt32 bytes)
{
	int status;

	if ((handle == NULL) || (pbuff == NULL))
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		RFM2G_STATUS result;

		result = RFM2gWrite(pdev->handle, offset, pbuff, bytes);
		if (result != RFM2G_SUCCESS)
		{
			ERR(RFM2gWrite, result);
			status = -1;
		}
		else
		{
			status = 0;
		}
	}
	return status;
}

/**	write 32-bit data to reflective memory
 *	@param[in]	handle	RM handle
 *	@param[in]	offset	offset
 *	@param[in]	data	data
 *	@retval	0	OK
 *	@retval	-1	API error
 *	@retval	-2	invalid argument
 */
int drvRM5565_write32(RM5565_t handle, epicsUInt32 offset, epicsUInt32 data)
{
	int status;

	if (handle == NULL)
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		RFM2G_STATUS result;

		result = RFM2gPoke32(pdev->handle, offset, data);
		if (result != RFM2G_SUCCESS)
		{
			ERR(RFM2gPoke32, result);
			status = -1;
		}
		else
		{
			status = 0;
		}
	}
	return status;
}

/**	write 16-bit data to reflective memory
 *	@param[in]	handle	RM handle
 *	@param[in]	offset	offset
 *	@param[in]	data	data
 *	@retval	0	OK
 *	@retval	-1	API error
 *	@retval	-2	invalid argument
 */
int drvRM5565_write16(RM5565_t handle, epicsUInt32 offset, epicsUInt16 data)
{
	int status;

	if (handle == NULL)
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		RFM2G_STATUS result;

		result = RFM2gPoke16(pdev->handle, offset, data);
		if (result != RFM2G_SUCCESS)
		{
			ERR(RFM2gPoke16, result);
			status = -1;
		}
		else
		{
			status = 0;
		}
	}
	return status;
}

/**	write 8-bit data to reflective memory
 *	@param[in]	handle	RM handle
 *	@param[in]	offset	offset
 *	@param[in]	data	data
 *	@retval	0	OK
 *	@retval	-1	API error
 *	@retval	-2	invalid argument
 */
int drvRM5565_write8(RM5565_t handle, epicsUInt32 offset, epicsUInt8 data)
{
	int status;

	if (handle == NULL)
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		RFM2G_STATUS result;

		result = RFM2gPoke8(pdev->handle, offset, data);
		if (result != RFM2G_SUCCESS)
		{
			ERR(RFM2gPoke8, result);
			status = -1;
		}
		else
		{
			status = 0;
		}
	}
	return status;
}

#if		defined(linux)
/**	map user memory
 *	@param[in]	handle	RM handle
 *	@param[out]	puser	pointer to user memory pointer
 *	@param[in]	offset	offset
 *	@param[in]	size	memory size
 *	@retval	0	OK
 *	@retval	-1	API error
 *	@retval	-2	invalid argument
 */
int drvRM5565_map(RM5565_t handle, volatile void **puser, rmoffset_t offset, epicsUInt32 size)
{
	int status;

	if (handle == NULL)
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		mapinfo_t *pmap;
		RFM2G_STATUS result;

		pmap = (mapinfo_t *)malloc(sizeof(mapinfo_t));
		if (pmap == NULL)
		{
			status = -1;
		}
		else
		{
			result = RFM2gUserMemoryBytes(pdev->handle, puser, offset, size);
			if (result != RFM2G_SUCCESS)
			{
				ERR(RFM2gUserMemoryBytes, result);
				free(pmap);
				status = -1;
			}
			else
			{
				pmap->memptr = *puser;
				pmap->offset = offset;
				pmap->size = size;
				ellAdd(&pdev->maplist, &pmap->link);
				status = 0;
			}
		}
	}
	return status;
}

/**	map user memory with DMA
 *	@param[in]	handle	RM handle
 *	@param[out]	puser	pointer to user memory pointer
 *	@param[in]	offset	offset
 *	@param[in]	size	memory size
 *	@retval	0	OK
 *	@retval	-1	API error
 *	@retval	-2	invalid argument
 */
int drvRM5565_mapDMA(RM5565_t handle, volatile void **puser, rmoffset_t offset, epicsUInt32 size)
{
	int status;

	if (handle == NULL)
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		mapinfo_t *pmap;
		RFM2G_STATUS result;

		pmap = (mapinfo_t *)malloc(sizeof(mapinfo_t));
		if (pmap == NULL)
		{
			status = -1;
		}
		else
		{
			offset |= RFM2G_DMA_MMAP_OFFSET;
			result = RFM2gUserMemoryBytes(pdev->handle, puser, offset, size);
			if (result != RFM2G_SUCCESS)
			{
				ERR(RFM2gUserMemoryBytes, result);
				free(pmap);
				status = -1;
			}
			else
			{
				pmap->memptr = *puser;
				pmap->offset = offset;
				pmap->size = size;
				ellAdd(&pdev->maplist, &pmap->link);
				status = 0;
			}
		}
	}
	return status;
}

/**	unmap user memory
 *	@param[in]	handle	RM handle
 *	@param[in]	puser	pointer to user memory pointer
 *	@param[in]	size	map size
 *	@retval	0	OK
 *	@retval	-1	API error
 *	@retval	-2	invalid argument
 */
int drvRM5565_unmap(RM5565_t handle, volatile void **puser, epicsUInt32 size)
{
	int status;

	if (handle == NULL)
	{
		(void)errlogPrintf(__FILE__":%d invalid argument\n", __LINE__);
		status = -2;
	}
	else
	{
		devinfo_t *pdev = (devinfo_t *)handle;
		mapinfo_t *pmap;

		status = -2;
		pmap = (mapinfo_t *)ellFirst(&pdev->maplist);
		while (pmap != NULL)
		{
			if ((pmap->memptr == *puser) && (pmap->size == size))
			{
				RFM2G_STATUS result;

				result = RFM2gUnMapUserMemoryBytes(pdev->handle, puser, size);
				if (result != RFM2G_SUCCESS)
				{
					ERR(RFM2gUnMapUserMemoryBytes, result);
					status = -1;
				}
				else
				{
					status = 0;
				}
				ellDelete(&pdev->maplist, &pmap->link);
				free(pmap);
				break;
			}
			pmap = (mapinfo_t *)ellNext(&pmap->link);
		}
	}
	return status;
}
#endif	/* defined(linux) */

/**	polling BAD DATA status
 *	@param[in]	pCB	pointer to CALLBACK structure
 */
static void poll_baddata(CALLBACK *pCB)
{
	devinfo_t *pdev;
	RFM2G_STATUS result;
	RFM2G_UINT32 lcsr;

	callbackGetUser(pdev, pCB);
	assert(pdev != NULL);

	API(RFM2gClearEvent, pdev->handle, RFM2GEVENT_BAD_DATA);
	epicsThreadSleep(0.5);
	result = RFM2gReadReg(pdev->handle, RFM2GCTRLREGMEM, rfm2g_lcsr, sizeof(lcsr), &lcsr);
	if (result != RFM2G_SUCCESS)
	{
		ERR(RFM2gReadReg, result);
		callbackRequestDelayed(&pdev->poll_baddata, 0.5);
	}
	else if ((lcsr & RFM2G_LCSR_BADDATA) != 0)
	{
		callbackRequestDelayed(&pdev->poll_baddata, 0.5);
	}
	else
	{
		scanIoRequest(pdev->scan_baddata);
		API(RFM2gEnableEvent, pdev->handle, RFM2GEVENT_BAD_DATA);
	}
}

/**	polling SYNC LOSS status
 *	@param[in]	pCB	pointer to CALLBACK structure
 */
static void poll_syncloss(CALLBACK *pCB)
{
	devinfo_t *pdev;
	RFM2G_STATUS result;
	RFM2G_UINT32 lcsr;

	callbackGetUser(pdev, pCB);
	assert(pdev != NULL);

	API(RFM2gClearEvent, pdev->handle, RFM2GEVENT_SYNC_LOSS);
	epicsThreadSleep(0.5);
	result = RFM2gReadReg(pdev->handle, RFM2GCTRLREGMEM, rfm2g_lcsr, sizeof(lcsr), &lcsr);
	if (result != RFM2G_SUCCESS)
	{
		ERR(RFM2gReadReg, result);
		callbackRequestDelayed(&pdev->poll_syncloss, 0.5);
	}
	else if ((lcsr & RFM2G_LCSR_SNCLOST) != 0)
	{
		callbackRequestDelayed(&pdev->poll_syncloss, 0.5);
	}
	else
	{
		scanIoRequest(pdev->scan_syncloss);
		API(RFM2gEnableEvent, pdev->handle, RFM2GEVENT_SYNC_LOSS);
	}
}

/**	generate driver report
 *	@param[in]	level	verbosity level
 *	@return	0
 */
static long drv_report(int level)
{
	devinfo_t *pdev;

	pdev = (devinfo_t *)ellFirst(&l_list);
	while (pdev != NULL)
	{
		RFM2G_UINT8 bid;
		RFM2G_NODE node;
		char name[64u], drvver[64u], dllver[64u];

		API(RFM2gDeviceName, pdev->handle, name);
		API(RFM2gBoardID, pdev->handle, &bid);
		API(RFM2gNodeID, pdev->handle, &node);

		if (level == 0)
		{
			(void)fprintf(stderr, "[%s] %d node:%d\n", name, bid, node);
		}
		else
		{
			if (level > 0)
			{
				API(RFM2gDriverVersion, pdev->handle, drvver);
				API(RFM2gDllVersion, pdev->handle, dllver);

				(void)fprintf(stderr, "Device Name : %s\n", name);
				(void)fprintf(stderr, "Version     : %s/%s (drv/dll)\n", drvver, dllver);
				(void)fprintf(stderr, "Board ID    : %d\n", bid);
				(void)fprintf(stderr, "Node ID     : %d\n", node);
				(void)fprintf(stderr, "Offset      : 0x%08X - 0x%08X\n", pdev->offset_sta, pdev->offset_end);
			}
			if (level > 1)
			{
				RFM2GCONFIG cnfg;

				API(RFM2gGetConfig, pdev->handle, &cnfg);

				(void)fprintf(stderr, "PLX rev.    : %d\n", cnfg.PlxRevision);
				(void)fprintf(stderr, "Board rev.  : %d\n", cnfg.BoardRevision);
				(void)fprintf(stderr, "Board build : %d\n", cnfg.BuildId);
				(void)fprintf(stderr, "LCSR1       : 0x%08X\n", cnfg.Lcsr1);
				(void)fprintf(stderr, "PCI rev. ID : %d\n", cnfg.RevisionId);
			}
		}
		pdev = (devinfo_t *)ellNext(&pdev->link);
	}
	return 0L;
}

/**	terminate driver
 *	@param[in]	arg	argument(but not used)
 */
static void drv_term(void *arg)
{
	(void)arg;
	devinfo_t *pdev;

	pdev = (devinfo_t *)ellLast(&l_list);
	if (pdev == NULL)
	{
		(void)errlogPrintf(__FILE__":%d device is not opened\n", __LINE__);
	}
	else
	{
		RFM2G_STATUS result;

		do
		{
			devinfo_t *pdevPrev = (devinfo_t *)ellPrevious(&pdev->link);

			disable_callback(pdev);

			delete_fifos(pdev);

			unmap_all(pdev);

			if ((result = RFM2gClose(&pdev->handle)) != RFM2G_SUCCESS)
			{
				ERR(RFM2gClose, result);
			}
			ellDelete(&l_list, &pdev->link);
			free(pdev);
			pdev = pdevPrev;
		}
		while (pdev != NULL);
	}
}

/**	initialize driver
 *	@retval	0	OK
 *	@retval	-1	NG
 */
static long drv_init(void)
{
	return 0L;
}

/**	@cond		*/
static const struct {
	long number;
	DRVSUPFUN report;
	DRVSUPFUN init;
} drvRM5565 = {
	2L,
	drv_report,
	drv_init
};
epicsExportAddress(drvet, drvRM5565);
/**	@endcond	*/

/**	event handler for RFM2G
 *	@param[in]	handle	RFM2G driver handle
 *	@param[in]	evtinfo	pointer to event information structure
 */
static void eventfunc(RFM2GHANDLE handle, RFM2GEVENTINFO *evtinfo)
{
	devinfo_t *pdev;

	pdev = (devinfo_t *)ellFirst(&l_list);
	while (pdev != NULL)
	{
		if (pdev->handle == handle)
		{
			break;
		}
		pdev = (devinfo_t *)ellNext(&pdev->link);
	}
	if (pdev != NULL)
	{
		epicsUInt8 sid;
		epicsUInt32 isd;

		switch (evtinfo->Event)
		{
		case RFM2GEVENT_INTR1:
			isd = evtinfo->ExtendedInfo;
			sid = evtinfo->NodeId;
			epicsRingBytesPut(pdev->isd[0u], (char *)&isd, sizeof(isd));
			epicsRingBytesPut(pdev->sid[0u], (char *)&sid, sizeof(sid));
			scanIoRequest(pdev->scan_nint[0u]);
			break;

		case RFM2GEVENT_INTR2:
			isd = evtinfo->ExtendedInfo;
			sid = evtinfo->NodeId;
			epicsRingBytesPut(pdev->isd[1u], (char *)&isd, sizeof(isd));
			epicsRingBytesPut(pdev->sid[1u], (char *)&sid, sizeof(sid));
			scanIoRequest(pdev->scan_nint[1u]);
			break;

		case RFM2GEVENT_INTR3:
			isd = evtinfo->ExtendedInfo;
			sid = evtinfo->NodeId;
			epicsRingBytesPut(pdev->isd[2u], (char *)&isd, sizeof(isd));
			epicsRingBytesPut(pdev->sid[2u], (char *)&sid, sizeof(sid));
			scanIoRequest(pdev->scan_nint[2u]);
			break;

		case RFM2GEVENT_INTR4:
			isd = evtinfo->ExtendedInfo;
			sid = evtinfo->NodeId;
			epicsRingBytesPut(pdev->isd[3u], (char *)&isd, sizeof(isd));
			epicsRingBytesPut(pdev->sid[3u], (char *)&sid, sizeof(sid));
			scanIoRequest(pdev->scan_nint[3u]);
			break;

		case RFM2GEVENT_BAD_DATA:
			pdev->count_baddata++;
			scanIoRequest(pdev->scan_baddata);
			callbackRequest(&pdev->poll_baddata);
			break;

		case RFM2GEVENT_RXFIFO_FULL:
			pdev->count_fifofull++;
			scanIoRequest(pdev->scan_fifofull);
			break;

		case RFM2GEVENT_ROGUE_PKT:
			pdev->count_roguedet++;
			scanIoRequest(pdev->scan_roguedet);
			break;

		case RFM2GEVENT_RXFIFO_AFULL:
			pdev->count_fifoafull++;
			scanIoRequest(pdev->scan_fifoafull);
			break;

		case RFM2GEVENT_SYNC_LOSS:
			pdev->count_syncloss++;
			scanIoRequest(pdev->scan_syncloss);
			callbackRequest(&pdev->poll_syncloss);
			break;

		default:
			break;
		}
	}
}

/**	configure driver of reflective memory
 *	@param[in]	card	card number @n
 *						任意で重複しない番号を指定する。 @n
 *						この番号は、EPICS DBファイル中で使用される。
 *	@param[in]	devpath	device path @n
 *						デバイスパス (/dev/rfm2g[0-9]) を指定する。 @n
 *						1枚のみ装着の場合は、通常 "/dev/rfm2g0" になる。
 */
static void RM5565PciConfigure(int card, char *devpath)
{
	static int s_flag = 0;
	devinfo_t *pdev;

	pdev = (devinfo_t *)ellFirst(&l_list);
	while (pdev != NULL)
	{
		if ((pdev->card == card) || (strcmp(pdev->devpath, devpath) == 0))
		{
			break;
		}
	}
	if (pdev != NULL)
	{
		(void)errlogPrintf(__FILE__":%d already registered\n", __LINE__);
	}
	else
	{
		RFM2GHANDLE handle = NULL;
		RFM2G_STATUS result;

		result = RFM2gOpen(devpath, &handle);
		if (result != RFM2G_SUCCESS)
		{
			ERR(RFM2gOpen, result);
		}
		else
		{
			size_t i;

			pdev = (devinfo_t *)mallocMustSucceed(sizeof(devinfo_t), __func__);
			memset(pdev, 0, sizeof(devinfo_t));

			pdev->card = card;
			strcpy(pdev->devpath, devpath);
			pdev->handle = handle;

			ellInit(&pdev->maplist);

			callbackSetCallback(poll_baddata, &pdev->poll_baddata);
			callbackSetUser(pdev, &pdev->poll_baddata);
			callbackSetPriority(priorityLow, &pdev->poll_baddata);

			callbackSetCallback(poll_syncloss, &pdev->poll_syncloss);
			callbackSetUser(pdev, &pdev->poll_syncloss);
			callbackSetPriority(priorityLow, &pdev->poll_syncloss);

			if (create_fifos(pdev) != 0)
			{
				(void)errlogPrintf(__FILE__":%d create_fifos() failed\n", __LINE__);
				delete_fifos(pdev);
				API(RFM2gClose, &pdev->handle);
				free(pdev);
			}
			else
			{
				RFM2G_UINT32 first, size;

				API(RFM2gFirst, pdev->handle, &first);
				API(RFM2gSize, pdev->handle, &size);
				pdev->offset_sta = first;
				pdev->offset_end = first + size - 1u;

				scanIoInit(&pdev->scan_baddata);
				scanIoInit(&pdev->scan_fifofull);
				scanIoInit(&pdev->scan_roguedet);
				scanIoInit(&pdev->scan_fifoafull);
				scanIoInit(&pdev->scan_syncloss);
				for (i = 0u; i < countof(pdev->scan_nint); i++)
				{
					scanIoInit(&pdev->scan_nint[i]);
				}

				enable_callback(pdev, eventfunc);

				ellAdd(&l_list, &pdev->link);
				if (__sync_bool_compare_and_swap(&s_flag, 0, 1))
				{
					epicsAtExit(drv_term, NULL);
				}
			}
		}
	}
}

/**
 *	structure of debug flag name table
 */
static const struct {
	const char *name;	/**< debug flag name		*/
	RFM2G_UINT32 flag;	/**< value of debug flag	*/
} l_dbtable[] = {
	{"all",      0xFFFFFFFFu},
	{"close",    RFM2G_DBCLOSE},
	{"dma",      RFM2G_DBDMA},
	{"error",    RFM2G_DBERROR},
	{"init",     RFM2G_DBINIT},
	{"intr",     RFM2G_DBINTR},
	{"ioctl",    RFM2G_DBIOCTL},
	{"minphys",  RFM2G_DBMINPHYS},
	{"mmap",     RFM2G_DBMMAP},
	{"mutex",    RFM2G_DBMUTEX},
	{"not_intr", RFM2G_DBINTR_NOT},
	{"open",     RFM2G_DBOPEN},
	{"peek",     RFM2G_DBPEEK},
	{"poke",     RFM2G_DBPOKE},
	{"read",     RFM2G_DBREAD},
	{"slow",     RFM2G_DBSLOW},
	{"strat",    RFM2G_DBSTRAT},
	{"timer",    RFM2G_DBTIMER},
	{"trace",    RFM2G_DBTRACE},
	{"write",    RFM2G_DBWRITE}
};

/**	validity check for debug flag name
 *	@param[in]	flagname	debug flag name
 *	@param[out]	pstat		pointer to boolean variable that stores set/clear condition
 *	@retval	-1		invalid flag name
 *	@retval	other	index of l_dbtable array
 */
static int validflag(const char *flagname, bool *pstat)
{
	const char *pstr;
	size_t i, matched;
	int match[20u];

	if (flagname[0u] == '-')
	{
		*pstat = false;
		pstr = &flagname[1u];
	}
	else
	{
		*pstat = true;
		pstr = flagname;
	}
	matched = 0u;
	for (i = 0u; i < countof(l_dbtable); i++)
	{
		if (epicsStrnCaseCmp(l_dbtable[i].name, pstr, strlen(pstr)) == 0)
		{
			match[matched++] = i;
		}
	}
	if (matched == 0u)
	{
		(void)fprintf(stderr, "invalid flag \'%s\'.\n", pstr);
		return -1;
	}
	else if (matched > 1u)
	{
		(void)fprintf(stderr, "ambiguous flag \'%s\'. this could be\n", pstr);
		for (i = 0u; i < matched; i++)
		{
			(void)fprintf(stderr, "  \'%s\'\n", l_dbtable[match[i]].name);
		}
		return -1;
	}
	return match[0u];
}

/**	set/clear debug flag
 *	@param[in]	card		card number
 *	@param[in]	flagname	debug flag name
 */
void RM5565DebugFlag(int card, const char *flagname)
{
	devinfo_t *pdev;

	if (flagname == NULL)
	{
		(void)fputs("flag name is not specified\n", stderr);
	}
	else if ((pdev = find_dev(card)) == NULL)
	{
		(void)fprintf(stderr, "invalid card number:%d\n", card);
	}
	else
	{
		bool stat;
		int index;

		index = validflag(flagname, &stat);
		if (index >= 0)
		{
			RFM2G_UINT32 flags;
			RFM2G_STATUS result;

			result = RFM2gGetDebugFlags(pdev->handle, &flags);
			if (result != RFM2G_SUCCESS)
			{
				ERR(RFM2gGetDebugFlags, result);
			}
			else
			{
				if (stat)
				{
					flags |= l_dbtable[index].flag;
				}
				else
				{
					flags &= ~l_dbtable[index].flag;
				}
				result = RFM2gSetDebugFlags(pdev->handle, flags);
				if (result != RFM2G_SUCCESS)
				{
					ERR(RFM2gSetDebugFlags, result);
				}
				else if (stat)
				{
					(void)fprintf(stderr, "set \'%s\' flag\n", l_dbtable[index].name);
				}
				else
				{
					(void)fprintf(stderr, "cleared \'%s\' flag\n", l_dbtable[index].name);
				}
			}
		}
	}
}

/**	@page cmd_RM5565PciConfigure RM5565PciConfigure
 *	RM5565PciConfigure(@e card_number, @e device_path)
 *	@param	card_number	Card number
 *	@param	device_path Device path
 *
 *	内部で @ref RM5565PciConfigure を呼び出す。
 *	パラメータの詳細については @ref RM5565PciConfigure "そちら" を参照のこと。
 */
/**	@cond		*/
static const iocshArg RM5565PciConfigure_arg0 = {"card_number", iocshArgInt};
static const iocshArg RM5565PciConfigure_arg1 = {"device_path", iocshArgString};
static const iocshArg * const RM5565PciConfigure_args[] = {
	&RM5565PciConfigure_arg0,
	&RM5565PciConfigure_arg1
};
static const iocshFuncDef RM5565PciConfigure_def = {
	"RM5565PciConfigure",
	2,
	RM5565PciConfigure_args
};
static void RM5565PciConfigure_caller(const iocshArgBuf args[])
{
	RM5565PciConfigure(args[0u].ival, args[1u].sval);
}
/**	@endcond	*/

/**	@page cmd_RM5565DebugFlag RM5565DebugFalg
 *	RM5565DebugFlag(@e card_number, @e debug_flag_name)
 *	@param	card_number	Card number @n
 *						RM5565PciConfigure の第一パラメータに指定した番号を指定する。
 *	@param	debug_flag_name	Name of debug-flag @n
 *						デバッグフラグ名を指定する。指定できるのは下記。
 *						Name     | Meaning
 *						:--------|:------------------------------------
 *						all      | all
 *						close    | Trace close(2) system calls
 *						dma      | Trace DMA calles
 *						error    | Report critical errors
 *						init     | Trace device probing and search
 *						intr     | Trace interrupt service
 *						ioctl    | Trace ioctl(2) system calls
 *						minphys  | Trace minphys limits
 *						mmap     | Trace mmap(2) system calls
 *						mutex    | Trace synchronization and locking
 *						not_intr | Trace non-RFM interrupt service
 *						open     | Trace open(2) system calls
 *						peek     | Trace peeks
 *						poke     | Trace pokes
 *						read     | Trace read(2) system calls
 *						slow     | Let syslogd get the message
 *						strat    | Trace read/write strategy
 *						timer    | Trace interval timer
 *						trace    | Trace subroutine entry/exit
 *						write    | Trace write(2) system calls
 *
 *	内部で @ref RM5565DebugFlag を呼び出す。
 */
/**	@cond		*/
static const iocshArg RM5565DebugFlag_arg0 = {"card_number", iocshArgInt};
static const iocshArg RM5565DebugFlag_arg1 = {"debug_flag_name", iocshArgString};
static const iocshArg * const RM5565DebugFlag_args[] = {
	&RM5565DebugFlag_arg0,
	&RM5565DebugFlag_arg1
};
static const iocshFuncDef RM5565DebugFlag_def = {
	"RM5565DebugFlag",
	2,
	RM5565DebugFlag_args
};
static void RM5565DebugFlag_caller(const iocshArgBuf args[])
{
	RM5565DebugFlag(args[0u].ival, args[1u].sval);
}
/**	@endcond	*/

/**	@cond		*/
/*
 *	registrar
 */
static void drvRM5565_registrar(void)
{
	iocshRegister(&RM5565PciConfigure_def, RM5565PciConfigure_caller);
	iocshRegister(&RM5565DebugFlag_def, RM5565DebugFlag_caller);
}
epicsExportRegistrar(drvRM5565_registrar);
/**	@endcond	*/

